import logging
logger = logging.getLogger(__name__)
import glob


class Anonymizer():
    def __init__(self):
        pass

    def file_anonymization(self, filename, output_filename=None):
        pass

    def recursive_anonymization(self, path, output_path=None):
        dirlist = glob.glob(path)

        pass
